////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2008, Lawrence Livermore National Security,
// LLC. Produced at the Lawrence Livermore National Laboratory. 
// Written by Erik Draeger, draeger1@llnl.gov.
//
// LLNL-CODE-432558 All rights reserved.  
//
// This file is part of bigFFT.
//
// bigFFT is distributed under the terms of the GNU General Public License
// as published by the Free Software Foundation, either version 2 of
// the License, or (at your option) any later version.
// See the file COPYING in the root directory of this distribution
// or <http://www.gnu.org/licenses/>.
////////////////////////////////////////////////////////////////////////////////
// 
// bigFFT:  real-to-complex 3D transform using a two dimensional process grid
//

//  bigfft_forward:   input is rhor[nx*ny*(nz+2)/npes], data is stored with x local 
//                    and varying fastest, i.e. grid point (x,y,z) is at: 
//                         rhor[x+y*bigdata->np0_+z*bigdata->np0_*bigdata->np1locf_]
//                    output is in place, using Hermitian symmetry rhok[i] = conj[rhok[N-i]] 
//                    data is stored with z local and varying fastest.
//
//  bigfft_backward:  input is rhor[nx*ny*(nz+2)/npes], complex data stored using Hermitian 
//                    symmetry.
//                    output by bigfft_forward.  Output is in place, and in same order as 
//                    input array of bigfft_forward.
//
// written by Erik Draeger, LLNL, 12/19/2008

#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <complex.h>
#include <assert.h>
#ifdef USE_FFTW
#ifndef __APPLE__
#include <fftw.h>
#else
#include <dfftw.h>
#endif
#endif
#include <mpi.h>
#include "bigFFT.h"
#include "ProcessGrid2D.h"
// ddcMD timing routine
//#include "ptiming.h"

#ifdef USE_ESSL
#include "essl.h"
#endif

////////////////////////////////////////////////////////////////////////////////
int bigfft_init(BIGFFT_DATA *bigdata, MPI_Comm comm, int np0, int np1, int np2, int nprow, int npcol) 
{
  bigdata->np0_ = np0;
  bigdata->np1_ = np1;
  bigdata->np2_ = np2;
  bigdata->nprow_ = nprow;
  bigdata->npcol_ = npcol;

#ifdef USE_FFTW
//ewd DEBUG:  use FFTW_ESTIMATE instead of FFTW_MEASURE for now to speed debugging
  bigdata->fwplan0 = fftw_create_plan(bigdata->np0_,FFTW_FORWARD,FFTW_ESTIMATE|FFTW_OUT_OF_PLACE);
  bigdata->fwplan1 = fftw_create_plan(bigdata->np1_,FFTW_FORWARD,FFTW_ESTIMATE|FFTW_OUT_OF_PLACE);
  bigdata->fwplan2 = fftw_create_plan(bigdata->np2_,FFTW_FORWARD,FFTW_ESTIMATE|FFTW_OUT_OF_PLACE);
  bigdata->bwplan0 = fftw_create_plan(bigdata->np0_,FFTW_BACKWARD,FFTW_ESTIMATE|FFTW_OUT_OF_PLACE);
  bigdata->bwplan1 = fftw_create_plan(bigdata->np1_,FFTW_BACKWARD,FFTW_ESTIMATE|FFTW_OUT_OF_PLACE);
  bigdata->bwplan2 = fftw_create_plan(bigdata->np2_,FFTW_BACKWARD,FFTW_ESTIMATE|FFTW_OUT_OF_PLACE);
#endif

  PGRID_DATA* pdata;
  pdata = calloc(1,sizeof(PGRID_DATA));
  pgrid_init(pdata,comm,bigdata->nprow_,bigdata->npcol_);
  bigdata->mype_ = pgrid_mype(pdata);
  bigdata->npes_ = pgrid_npes(pdata);
  bigdata->myrow_ = pgrid_myrow(pdata,bigdata->mype_);
  bigdata->mycol_ = pgrid_mycol(pdata,bigdata->mype_);

  // calculate local data sizes 
  bigdata->np0loc_row_ = (int*)malloc(sizeof(int)*bigdata->nprow_);
  bigdata->np1loc_row_ = (int*)malloc(sizeof(int)*bigdata->nprow_);
  bigdata->np1loc_col_ = (int*)malloc(sizeof(int)*bigdata->npcol_);
  bigdata->np2loc_col_ = (int*)malloc(sizeof(int)*bigdata->npcol_);
  bigdata->iprow_y_ = (int*)malloc(sizeof(int)*bigdata->np1_);
  bigdata->ipcol_z_ = (int*)malloc(sizeof(int)*bigdata->np2_);

  { // limit scope
     int np0half = np0/2 + 1;
     bigdata->np0half_ = np0half;
     int np0loc = np0half/nprow;
     for (int i=0; i<nprow; i++)
	bigdata->np0loc_row_[i] = np0loc;
     for (int i=0; i<np0half%nprow; ++i)
	++bigdata->np0loc_row_[i];
     bigdata->np0loc_ = bigdata->np0loc_row_[bigdata->myrow_];    
  } // limit scope
  
  { //limit scope
     int np1loc = np1/nprow;
     for (int i=0; i<nprow; i++)
	bigdata->np1loc_row_[i] = np1loc;
     for (int i=0; i<np1%nprow; i++)
	++bigdata->np1loc_row_[i];
     bigdata->np1locf_ = bigdata->np1loc_row_[bigdata->myrow_];    
  } // limit scope


  { //limit scope
     int np1loc = np1/npcol;
     for (int i=0; i<npcol; i++)
	bigdata->np1loc_col_[i] = np1loc;
     for (int i=0; i<np1%npcol; i++)
	++bigdata->np1loc_col_[i];
    bigdata->np1locb_ = bigdata->np1loc_col_[bigdata->mycol_];    
  } // limit scope

  { //limit scope
     int np2loc = np2/npcol;
     for (int i=0; i<npcol; i++)
	bigdata->np2loc_col_[i] = np2loc;
     for (int i=0; i<np2%npcol; i++)
	++bigdata->np2loc_col_[i];
     bigdata->np2loc_ = bigdata->np2loc_col_[bigdata->mycol_];    
  } // limit scope

  int iy = 0;
  for (int i=0; i<bigdata->nprow_; i++) 
  {  
    const int np1loc = bigdata->np1loc_row_[i];
    for (int j=0; j<np1loc; j++) 
    {
      bigdata->iprow_y_[iy] = i;
      ++iy;
    }
  }
  int iz = 0;
  for (int i=0; i<bigdata->npcol_; i++) 
  {  
    const int np2loc = bigdata->np2loc_col_[i];
    for (int j=0; j<np2loc; j++) 
    {
      bigdata->ipcol_z_[iz] = i;
      ++iz;
    }
  }

  bigdata->comm_ = pgrid_comm(pdata);
  // create single-row communicators w. npcol tasks
  for (int i=0; i<bigdata->nprow_; i++) 
  {
    MPI_Comm tmpcomm = pgrid_subcomm(pdata,1,bigdata->npcol_,i,0);
    if (i == bigdata->myrow_)
      bigdata->rowcomm_ = tmpcomm;
    else if (tmpcomm != 0 && tmpcomm != MPI_COMM_NULL)
      MPI_Comm_free(&tmpcomm);
  }
 
  // create single-column communicators w. nprow tasks
  for (int j=0; j<bigdata->npcol_; j++) 
  {
    MPI_Comm tmpcomm = pgrid_subcomm(pdata,bigdata->nprow_,1,0,j);
    if (j == bigdata->mycol_)
      bigdata->colcomm_ = tmpcomm;
    else if (tmpcomm != 0 && tmpcomm != MPI_COMM_NULL)
      //else if (tmpcomm != 0)
      MPI_Comm_free(&tmpcomm);
  } 

  // allocate work spaces
  int size1 = np0*bigdata->np1locf_*bigdata->np2loc_;
  int size2 = np1*bigdata->np0_*bigdata->np2loc_/bigdata->nprow_;
  int size3 = np2*bigdata->np1locb_*bigdata->np0_/bigdata->nprow_;
  int maxsize = (size1 > size2 ? size1 : size2);
  maxsize = (maxsize > size3 ? maxsize : size3);
  bigdata->pack = (FftDataType*)malloc(sizeof(FftDataType)*maxsize);
  bigdata->fftin = (FftDataType*)malloc(sizeof(FftDataType)*maxsize);
  bigdata->fftout = (FftDataType*)malloc(sizeof(FftDataType)*maxsize);
  bigdata->mpibuf = (FftDataType*)malloc(sizeof(FftDataType)*maxsize);

#ifdef USE_ESSL
  // initialize pointers
  bigdata->aux1xf = 0; 
  bigdata->aux1yf = 0;
  bigdata->aux1zf = 0;
  bigdata->aux1xb = 0;
  bigdata->aux1yb = 0;  
  bigdata->aux1zb = 0;

  int na1x = (int) (20000 + 2.28 * np0);
  int na1y = (int) (20000 + 2.28 * np1);
  int na1z = (int) (20000 + 2.28 * np2);
  bigdata->naux1x = na1x;
  bigdata->naux1y = na1y;
  bigdata->naux1z = na1z;
  bigdata->aux1xf = (double*)malloc(sizeof(double)*na1x);
  bigdata->aux1yf = (double*)malloc(sizeof(double)*na1y);
  bigdata->aux1zf = (double*)malloc(sizeof(double)*na1z);
  bigdata->aux1xb = (double*)malloc(sizeof(double)*na1x);
  bigdata->aux1yb = (double*)malloc(sizeof(double)*na1y);
  bigdata->aux1zb = (double*)malloc(sizeof(double)*na1z);


  int naux2x = (int) (20000 + 2.28 * np0 + (256 + 2*np0)*64);
  bigdata->naux2 = naux2x;
  int naux2y = (int) (20000 + 2.28 * np1 + (256 + 2*np1)*64);
  bigdata->naux2 = ( bigdata->naux2 > naux2y ? bigdata->naux2 : naux2y );
  int naux2z = (int) (20000 + 2.28 * np2 + (256 + 2*np2)*64);
  bigdata->naux2 = ( bigdata->naux2 > naux2z ? bigdata->naux2 : naux2z );
  bigdata->aux2 = (double*)malloc(sizeof(double)*bigdata->naux2);

  // initialize x, y and z transforms
  int initflag = 1, inc1, inc2, isign;
  FftCommType scale = 1.0;
  FftDataType *p = 0;

  // x transforms
  inc1 = 1; inc2 = np0;
  isign = -1;
  int nxt = bigdata->np2loc_*bigdata->np1locf_;
#ifdef USE_SINGLEPREC_FFT
  //  scft(&initflag,p,&inc1,&inc2,p,&inc1,&inc2,&np0,&nxt,
  //      &isign,&scale,&bigdata->aux1xb[0],&bigdata->naux1x,&bigdata->aux2[0],&bigdata->naux2);
  scft(initflag,(void*)p,inc1,inc2,(void*)p,inc1,inc2,np0,nxt,
        isign,scale,&bigdata->aux1xb[0],bigdata->naux1x,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,p,inc1,inc2,p,inc1,inc2,np0,nxt,
        isign,scale,&bigdata->aux1xb[0],bigdata->naux1x,&bigdata->aux2[0],bigdata->naux2);
#endif
  isign = 1;
  int ntrans = bigdata->np1locf_*bigdata->np2loc_;
  bool ntrans_even = (ntrans%2 == 0);
  nxt = (ntrans_even ? ntrans/2 : ntrans/2 + 1);
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,(void*)p,inc1,inc2,(void*)p,inc1,inc2,np0,nxt,
        isign,scale,&bigdata->aux1xf[0],bigdata->naux1x,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,p,inc1,inc2,p,inc1,inc2,np0,nxt,
        isign,scale,&bigdata->aux1xf[0],bigdata->naux1x,&bigdata->aux2[0],bigdata->naux2);
#endif
  // y transforms
  inc1 = 1; inc2 = np1; ntrans = bigdata->np0loc_*bigdata->np2loc_;
  isign = -1;
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,(void*)p,inc1,inc2,(void*)p,inc1,inc2,np1,ntrans,
        isign,scale,&bigdata->aux1yb[0],bigdata->naux1y,&bigdata->aux2[0],bigdata->naux2);
  isign = 1;
  scft(initflag,(void*)p,inc1,inc2,(void*)p,inc1,inc2,np1,ntrans,
        isign,scale,&bigdata->aux1yf[0],bigdata->naux1y,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,p,inc1,inc2,p,inc1,inc2,np1,ntrans,
        isign,scale,&bigdata->aux1yb[0],bigdata->naux1y,&bigdata->aux2[0],bigdata->naux2);
  isign = 1;
  dcft(initflag,p,inc1,inc2,p,inc1,inc2,np1,ntrans,
        isign,scale,&bigdata->aux1yf[0],bigdata->naux1y,&bigdata->aux2[0],bigdata->naux2);
#endif  
  // z transforms
  inc1 = 1; inc2 = np2; ntrans = bigdata->np0loc_*bigdata->np1locb_;
  isign = -1;
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,(void*)p,inc1,inc2,(void*)p,inc1,inc2,np2,ntrans,
        isign,scale,&bigdata->aux1zb[0],bigdata->naux1z,&bigdata->aux2[0],bigdata->naux2);
  isign = 1;
  ntrans = bigdata->np0loc_*bigdata->np1locb_;
  scft(initflag,(void*)p,inc1,inc2,(void*)p,inc1,inc2,np2,ntrans,
        isign,scale,&bigdata->aux1zf[0],bigdata->naux1z,&bigdata->aux2[0],bigdata->naux2);  
#else
  dcft(initflag,p,inc1,inc2,p,inc1,inc2,np2,ntrans,
        isign,scale,&bigdata->aux1zb[0],bigdata->naux1z,&bigdata->aux2[0],bigdata->naux2);
  isign = 1;
  ntrans = bigdata->np0loc_*bigdata->np1locb_;
  dcft(initflag,p,inc1,inc2,p,inc1,inc2,np2,ntrans,
        isign,scale,&bigdata->aux1zf[0],bigdata->naux1z,&bigdata->aux2[0],bigdata->naux2);  
#endif  

#endif
  
  return 0;
}
////////////////////////////////////////////////////////////////////////////////
int bigfft_info_init(BIGFFT_DATA *bigdata, int np0, int np1, int np2, int nprow, int npcol) 
{
  bigdata->np0_ = np0;
  bigdata->np1_ = np1;
  bigdata->np2_ = np2;
  bigdata->nprow_ = nprow;
  bigdata->npcol_ = npcol;

  // calculate local data sizes 
  bigdata->np0loc_row_ = (int*)malloc(sizeof(int)*bigdata->nprow_);
  bigdata->np1loc_row_ = (int*)malloc(sizeof(int)*bigdata->nprow_);
  bigdata->np1loc_col_ = (int*)malloc(sizeof(int)*bigdata->npcol_);
  bigdata->np2loc_col_ = (int*)malloc(sizeof(int)*bigdata->npcol_);
  bigdata->iprow_y_ = (int*)malloc(sizeof(int)*bigdata->np1_);
  bigdata->ipcol_z_ = (int*)malloc(sizeof(int)*bigdata->np2_);

  { // limit scope
     int np0half = np0/2 + 1;
     bigdata->np0half_ = np0half;
     int np0loc = np0half/nprow;
     for (int i=0; i<nprow; i++)
	bigdata->np0loc_row_[i] = np0loc;
     for (int i=0; i<np0half%nprow; ++i)
	++bigdata->np0loc_row_[i];
     bigdata->np0loc_ = 0; //  myrow_ undefined bigdata->np0loc_row_[bigdata->myrow_];    
  } // limit scope
  
  { //limit scope
     int np1loc = np1/nprow;
     for (int i=0; i<nprow; i++)
	bigdata->np1loc_row_[i] = np1loc;
     for (int i=0; i<np1%nprow; i++)
	++bigdata->np1loc_row_[i];
     bigdata->np1locf_ = 0; // myrow_ undefined bigdata->np1loc_row_[bigdata->myrow_];    
  } // limit scope


  { //limit scope
     int np1loc = np1/npcol;
     for (int i=0; i<npcol; i++)
	bigdata->np1loc_col_[i] = np1loc;
     for (int i=0; i<np1%npcol; i++)
	++bigdata->np1loc_col_[i];
     bigdata->np1locb_ = 0; // mycol_ undefined bigdata->np1loc_col_[bigdata->mycol_];    
  } // limit scope

  { //limit scope
     int np2loc = np2/npcol;
     for (int i=0; i<npcol; i++)
	bigdata->np2loc_col_[i] = np2loc;
     for (int i=0; i<np2%npcol; i++)
	++bigdata->np2loc_col_[i];
     bigdata->np2loc_ = 0; // mycol_ undefined bigdata->np2loc_col_[bigdata->mycol_];    
  } // limit scope

  int iy = 0;
  for (int i=0; i<bigdata->nprow_; i++) 
  {  
    const int np1loc = bigdata->np1loc_row_[i];
    for (int j=0; j<np1loc; j++) 
    {
      bigdata->iprow_y_[iy] = i;
      ++iy;
    }
  }
  int iz = 0;
  for (int i=0; i<bigdata->npcol_; i++) 
  {  
    const int np2loc = bigdata->np2loc_col_[i];
    for (int j=0; j<np2loc; j++) 
    {
      bigdata->ipcol_z_[iz] = i;
      ++iz;
    }
  }

#ifdef USE_ESSL
  // initialize pointers
  bigdata->aux1xf = 0; 
  bigdata->aux1yf = 0;
  bigdata->aux1zf = 0;
  bigdata->aux1xb = 0;
  bigdata->aux1yb = 0;  
  bigdata->aux1zb = 0;
#endif
  
  return 0;
}
////////////////////////////////////////////////////////////////////////////////
void bigfft_cleanup(BIGFFT_DATA *bigdata) 
{
#ifdef USE_FFTW
  fftw_destroy_plan(bigdata->fwplan0);
  fftw_destroy_plan(bigdata->fwplan1);
  fftw_destroy_plan(bigdata->fwplan2);
  fftw_destroy_plan(bigdata->bwplan0);
  fftw_destroy_plan(bigdata->bwplan1);
  fftw_destroy_plan(bigdata->bwplan2);
#endif
  free(bigdata->pack);
  free(bigdata->fftin);
  free(bigdata->fftout);
  free(bigdata->mpibuf);

  free(bigdata->np0loc_row_);
  free(bigdata->np1loc_row_);
  free(bigdata->np1loc_col_);
  free(bigdata->np2loc_col_);

  //print_timings();
}

////////////////////////////////////////////////////////////////////////////////
int bigfft_forward(BIGFFT_DATA *bigdata, double* rhor, double scale) 
{

  // Goal:  want to transform real-space 3D density grid (real) to
  // reciprocal-space density (complex)
  //
  // Strategy: distribute data on 2D process grid, with x fully local and
  // y,z block distributed.
  //
  // input data needs to be in np0*np1loc*np2loc format, w. x varying fastest

  //ewd DEBUG
  //int mype = bigdata->mype_;
  //double t1 = MPI_Wtime();

  const int np0 = bigdata->np0_;
  const int np1 = bigdata->np1_;
  const int np2 = bigdata->np2_;
  const int np0half = bigdata->np0half_;
  int np0loc = bigdata->np0loc_;
  int np1locf = bigdata->np1locf_;
  int np1locb = bigdata->np1locb_;
  int np2loc = bigdata->np2loc_;
  const int nprow = bigdata->nprow_;
  const int npcol = bigdata->npcol_;


  // do two real transforms at a time, using Hermitian symmetry
  // do np1locf*np2loc 1D real FFTs in x-dir, two at a time
  int ntrans = np1locf*np2loc;
  bool ntrans_even = (ntrans%2 == 0);

  for (int k=0; k<ntrans/2; k++)
  {
    for (int i=0; i<np0; i++)
    {
      // cast rhor manually to ensure single precision case works
      double t1 = rhor[2*k*np0+i];
      double t2 = rhor[(2*k+1)*np0+i];
      const FftCommType v1 = (const FftCommType)t1;
      const FftCommType v2 = (const FftCommType)t2;
      bigdata->fftin[k*np0+i] = v1 + I*v2;
    }
  }
  if (!ntrans_even)
  {
    int kk = ntrans/2;
    for (int i=0; i<np0; i++)
    {
      double t1 = rhor[2*kk*np0+i];
      const FftCommType v1 = (const FftCommType)t1;
      bigdata->fftin[kk*np0+i] = v1 + I*0.0;
    }
  }


  //ewd DEBUG
  //double t2 = MPI_Wtime();
  //printf("bigFFT.forward time:  mype = %i, t2 time = %0.7f sec\n",mype,t2-t1);

  int inc1 = 1;
  int inc2 = np0;
  int nxt = (ntrans_even ? ntrans/2 : ntrans/2 + 1);
#ifdef USE_FFTW
  fftw(bigdata->fwplan0,nxt,(FFTW_COMPLEX*)&bigdata->fftin[0],inc1,inc2,
       (FFTW_COMPLEX*)&bigdata->fftout[0],inc1,inc2);
#endif
#ifdef USE_ESSL
  int initflag = 0;
  int isign = 1;
  FftCommType esslscale = 1.0;
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np0,nxt,isign,esslscale,&bigdata->aux1xf[0],bigdata->naux1x,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np0,nxt,isign,esslscale,&bigdata->aux1xf[0],bigdata->naux1x,&bigdata->aux2[0],bigdata->naux2);
#endif  
#endif  
  //ewd DEBUG
  //double t3 = MPI_Wtime();
  //printf("bigFFT.forward time:  mype = %i, t3 time = %0.7f sec\n",mype,t3-t2);


//ewd DEBUG DEBUG
//  if (mype == 3) {
//    int istart = 5*np0;
//    printf("DEBUG FWDX, fftout[0] = %0.12f, fftout[1] = %0.12f, fftout[2] = %0.12f, fftout[3] = %0.12f\n",creal(bigdata->fftout[istart]),creal(bigdata->fftout[istart+1]),creal(bigdata->fftout[istart+2]),creal(bigdata->fftout[istart+3]));
//    printf("DEBUG FWDX, fftout[np0] = %0.12f, fftout[np0-1] = %0.12f, fftout[np0-2] = %0.12f, fftout[np0-3] = %0.12f\n",creal(bigdata->fftout[istart+np0]),creal(bigdata->fftout[istart+np0-1]),creal(bigdata->fftout[istart+np0-2]),creal(bigdata->fftout[istart+np0-3]));
//  }

  // untangle data
  for (int k=0; k<ntrans/2; k++)
  {
    bigdata->pack[2*k*np0half] = creal(bigdata->fftout[k*np0]);
    bigdata->pack[(2*k+1)*np0half] = cimag(bigdata->fftout[k*np0]);
    for (int i=1; i<np0half; i++)
    {
      const FftCommType ti_re = 0.5*creal(bigdata->fftout[k*np0+i]);
      const FftCommType ti_im = 0.5*cimag(bigdata->fftout[k*np0+i]);
      const FftCommType tni_re = 0.5*creal(bigdata->fftout[k*np0+np0-i]);
      const FftCommType tni_im = 0.5*cimag(bigdata->fftout[k*np0+np0-i]);
      bigdata->pack[2*k*np0half+i] = ti_re+tni_re + I*(ti_im-tni_im);
      bigdata->pack[(2*k+1)*np0half+i] = ti_im+tni_im + I*(tni_re-ti_re);
    }
  }
  if (!ntrans_even)
  {
    int kk = ntrans/2;
    for (int i=0; i<np0half; i++)
      bigdata->pack[2*kk*np0half+i] = bigdata->fftout[kk*np0+i];
  }

  // rearrange data so that x will be distributed and y will end up local after
  // alltoallv i.e. current local index (x) must be varying slowest in sent array
  for (int k=0; k<np2loc; k++) 
    for (int j=0; j<np1locf; j++) 
      for (int i=0; i<np0half; i++) 
      {
        const FftDataType v = bigdata->pack[i+j*np0half+k*np0half*np1locf];
        bigdata->mpibuf[j+k*np1locf+i*np1locf*np2loc] = v;
      }

  // transpose
  int scnts[nprow],rcnts[nprow],sdisp[nprow],rdisp[nprow]; 
  int soffset = 0;
  int roffset = 0;
  for (int i=0; i<nprow; i++) 
  {
    int sendblk = 2*bigdata->np0loc_row_[i]*np1locf*np2loc;
    int recvblk = 2*np0loc*bigdata->np1loc_row_[i]*np2loc;
    scnts[i] = sendblk;
    rcnts[i] = recvblk;
    sdisp[i] = soffset;
    rdisp[i] = roffset;
    soffset += sendblk;
    roffset += recvblk;
  }

  //ewd DEBUG
  //double t4 = MPI_Wtime();
  //printf("bigFFT.forward time:  mype = %i, t4 time = %0.7f sec\n",mype,t4-t3);

  int status = MPI_Alltoallv((FftCommType*)&bigdata->mpibuf[0],&scnts[0],&sdisp[0],FFT_COMM_TYPE,
                             (FftCommType*)&bigdata->pack[0],&rcnts[0],&rdisp[0],FFT_COMM_TYPE,bigdata->colcomm_);
  assert ( status == 0 );

  //ewd DEBUG
  //double t5 = MPI_Wtime();
  //printf("bigFFT.forward time:  mype = %i, t5 time = %0.7f sec, alltoallv fwd1\n",mype,t5-t4);

  // reorder data so that y varies fastest for y-transform
  for (int i=0; i<np0loc; i++) 
    for (int k=0; k<np2loc; k++) 
    {
      int j = 0;
      int recvstart = 0;
      for (int tr=0; tr<nprow; tr++) 
      {
        for (int jloc=0; jloc<bigdata->np1loc_row_[tr]; jloc++) 
        {
          const FftDataType v = bigdata->pack[recvstart + k*bigdata->np1loc_row_[tr] + i*bigdata->np1loc_row_[tr]*np2loc + jloc];
          bigdata->fftin[j+k*np1+i*np1*np2loc] = v;
          j++;
        }
        recvstart += np0loc*bigdata->np1loc_row_[tr]*np2loc;
      }
    }

  inc2 = np1;
  int nyt = np0loc*np2loc;
#ifdef USE_FFTW
  fftw(bigdata->fwplan1,nyt,(FFTW_COMPLEX*)&bigdata->fftin[0],inc1,inc2,
       (FFTW_COMPLEX*)&bigdata->fftout[0],inc1,inc2);
#endif 
#ifdef USE_ESSL
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np1,nyt,isign,esslscale,&bigdata->aux1yf[0],bigdata->naux1y,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np1,nyt,isign,esslscale,&bigdata->aux1yf[0],bigdata->naux1y,&bigdata->aux2[0],bigdata->naux2);
#endif  
#endif  
  //ewd DEBUG
  //double t6 = MPI_Wtime();
  //printf("bigFFT.forward time:  mype = %i, t6 time = %0.7f sec\n",mype,t6-t5);


  // rearrange data so that y will be distributed and z will end up
  // local after alltoallv i.e. current local index (y) must be
  // varying slowest in sent array
  int nzt = np0loc*np2loc;
  for (int k=0; k<np2loc; k++) 
    for (int i=0; i<np0loc; i++) 
      for (int j=0; j<np1; j++) 
      {
        const FftDataType v = bigdata->fftout[j+k*np1+i*np1*np2loc];
        bigdata->mpibuf[k+i*np2loc+j*np2loc*np0loc] = v;
      }

  int scnts2[npcol],rcnts2[npcol],sdisp2[npcol],rdisp2[npcol]; 
  soffset = 0;
  roffset = 0;
  for (int i=0; i<npcol; i++) 
  {
    int sendblk = 2*np0loc*bigdata->np1loc_col_[i]*np2loc;
    int recvblk = 2*np0loc*np1locb*bigdata->np2loc_col_[i];
    scnts2[i] = sendblk;
    rcnts2[i] = recvblk;
    sdisp2[i] = soffset;
    rdisp2[i] = roffset;
    soffset += sendblk;
    roffset += recvblk;
  }

   //ewd DEBUG
  //double t7a = MPI_Wtime();
  //printf("bigFFT.forward time:  mype = %i, t7a time = %0.7f sec\n",mype,t7a-t6);

  status = MPI_Alltoallv((FftCommType*)&bigdata->mpibuf[0],&scnts2[0],&sdisp2[0],FFT_COMM_TYPE,
                         (FftCommType*)&bigdata->pack[0],&rcnts2[0],&rdisp2[0],FFT_COMM_TYPE,bigdata->rowcomm_);
  assert ( status == 0 );

  //ewd DEBUG
  //double t7b = MPI_Wtime();
  //printf("bigFFT.forward time:  mype = %i, t7b time = %0.7f sec, alltoallv fwd2 on %i procs\n",mype,t7b-t7a,npcol);

  // reorder data so that z varies fastest for z-transform
  for (int i=0; i<np0loc; i++) 
    for (int j=0; j<np1locb; j++) 
    {
      int k = 0;
      int recvstart = 0;
      for (int tc=0; tc<npcol; tc++) 
      {
        for (int kloc=0; kloc<bigdata->np2loc_col_[tc]; kloc++) 
        {
          const FftDataType v = bigdata->pack[recvstart + i*bigdata->np2loc_col_[tc] + j*bigdata->np2loc_col_[tc]*np0loc + kloc];
          bigdata->fftin[k+j*np2+i*np2*np1locb] = v;
          k++;
        }
        recvstart += np0loc*np1locb*bigdata->np2loc_col_[tc];
      }
    }

  inc2 = np2;
  nzt = np0loc*np1locb;
#ifdef USE_FFTW
  fftw(bigdata->fwplan2,nzt,(FFTW_COMPLEX*)&bigdata->fftin[0],inc1,inc2,
       (FFTW_COMPLEX*)&bigdata->fftout[0],inc1,inc2);
#endif 
#ifdef USE_ESSL
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np2,nzt,isign,esslscale,&bigdata->aux1zf[0],bigdata->naux1z,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np2,nzt,isign,esslscale,&bigdata->aux1zf[0],bigdata->naux1z,&bigdata->aux2[0],bigdata->naux2);
#endif  
#endif  
  //ewd DEBUG
  //double t8 = MPI_Wtime();
  //printf("bigFFT.forward time:  mype = %i, t8 time = %0.7f sec\n",mype,t8-t7b);

  for (int i=0; i<np2*np1locb*np0loc; i++) 
  {
    rhor[2*i] = (double)(scale*creal(bigdata->fftout[i]));
    rhor[2*i+1] = (double)(scale*cimag(bigdata->fftout[i]));
  }

  return 0;
}

////////////////////////////////////////////////////////////////////////////////
int bigfft_backward(BIGFFT_DATA *bigdata, double* rhok, double scale) 
{
  // Goal:  want to transform reciprocal-space 3D density grid (complex) to
  // real-space density (real)
  //
  // Strategy: distribute data on 2D process grid, with z fully local and
  // x,y block distributed.
  //
  // input data needs to be in np2*np1locf*np0loc format, w. z varying fastest

  //ewd DEBUG
  //int mype = bigdata->mype_;
  //double t1 = MPI_Wtime();

  const int np0 = bigdata->np0_;
  const int np1 = bigdata->np1_;
  const int np2 = bigdata->np2_;
  int np0loc = bigdata->np0loc_;
  int np1locf = bigdata->np1locf_;
  int np1locb = bigdata->np1locb_;
  int np2loc = bigdata->np2loc_;
  const int nprow = bigdata->nprow_;
  const int npcol = bigdata->npcol_;

  // copy rhok into fftout
  for (int i=0; i<np2*np1locb*np0loc; i++) 
  {
    double t1 = rhok[2*i];
    double t2 = rhok[2*i+1];
    const FftCommType re = (const FftCommType)t1;
    const FftCommType im = (const FftCommType)t2;
    bigdata->fftin[i] = re + I*im;
  }


  //ewd DEBUG
  //double t2 = MPI_Wtime();
  //printf("bigFFT.backward time:  mype = %i, t2 time = %0.7f sec\n",mype,t2-t1);

  // transform in z-direction for all local x and y
  int inc1 = 1;
  int inc2 = np2;
  int ntrans = np0loc*np1locb;
#ifdef USE_FFTW
  fftw(bigdata->bwplan2,ntrans,(FFTW_COMPLEX*)&bigdata->fftin[0],inc1,inc2,
       (FFTW_COMPLEX*)&bigdata->fftout[0],inc1,inc2);
#endif
#ifdef USE_ESSL
  int initflag = 0;
  int isign = -1;
  FftCommType esslscale = 1.0;
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np2,ntrans,isign,esslscale,&bigdata->aux1zb[0],bigdata->naux1z,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np2,ntrans,isign,esslscale,&bigdata->aux1zb[0],bigdata->naux1z,&bigdata->aux2[0],bigdata->naux2);
#endif  
#endif  

  //ewd DEBUG
  //double t3 = MPI_Wtime();
  //printf("bigFFT.backward time:  mype = %i, t3 time = %0.7f sec\n",mype,t3-t2);

  // rearrange data so that z will be distributed and y will end up local after
  // alltoallv i.e. current local index (z) must be varying slowest in sent array
  for (int i=0; i<np0loc; i++)
    for (int j=0; j<np1locb; j++)
      for (int k=0; k<np2; k++)
      {
        const FftDataType v = bigdata->fftout[k+j*np2+i*np2*np1locb];
        bigdata->mpibuf[j+i*np1locb+k*np1locb*np0loc] = v;
      }

  int scnts2[npcol],rcnts2[npcol],sdisp2[npcol],rdisp2[npcol]; 
  int soffset = 0;
  int roffset = 0;
  for (int i=0; i<npcol; i++) 
  {
    int sendblk = 2*np0loc*np1locb*bigdata->np2loc_col_[i];
    int recvblk = 2*np0loc*bigdata->np1loc_col_[i]*np2loc;
    scnts2[i] = sendblk;
    rcnts2[i] = recvblk;
    sdisp2[i] = soffset;
    rdisp2[i] = roffset;
    soffset += sendblk;
    roffset += recvblk;
  }

  //double t4 = MPI_Wtime();
  //printf("bigFFT.backward time:  mype = %i, t4 time = %0.7f sec\n",mype,t4-t3);

  int status = MPI_Alltoallv((FftCommType*)&bigdata->mpibuf[0],&scnts2[0],&sdisp2[0],FFT_COMM_TYPE,
                             (FftCommType*)&bigdata->pack[0],&rcnts2[0],&rdisp2[0],FFT_COMM_TYPE,bigdata->rowcomm_);
  assert ( status == 0 );

  //ewd DEBUG
  //double t5 = MPI_Wtime();
  //printf("bigFFT.backward time:  mype = %i, t5 time = %0.7f sec, alltoallv bwd1 on %i procs\n",mype,t5-t4,npcol);

  // reorder data so that y varies fastest for y-transform
  for (int k=0; k<np2loc; k++)
    for (int i=0; i<np0loc; i++)
    {
      int j = 0;
      int recvstart = 0;
      for (int tc=0; tc<npcol; tc++)
      {
        for (int jloc=0; jloc<bigdata->np1loc_col_[tc]; jloc++)
        {
          const FftDataType v = bigdata->pack[recvstart + i*bigdata->np1loc_col_[tc] + k*bigdata->np1loc_col_[tc]*np0loc + jloc];
          bigdata->fftin[j+i*np1+k*np1*np0loc] = v;
          j++;
        }
        recvstart += np0loc*bigdata->np1loc_col_[tc]*np2loc;
      }
    }

  inc2 = np1;
  int nyt = np0loc*np2loc;
#ifdef USE_FFTW
  fftw(bigdata->bwplan1,nyt,(FFTW_COMPLEX*)&bigdata->fftin[0],inc1,inc2,
       (FFTW_COMPLEX*)&bigdata->fftout[0],inc1,inc2);
#endif
#ifdef USE_ESSL
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np1,nyt,isign,esslscale,&bigdata->aux1yb[0],bigdata->naux1y,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np1,nyt,isign,esslscale,&bigdata->aux1yb[0],bigdata->naux1y,&bigdata->aux2[0],bigdata->naux2);
#endif  
#endif  

  //ewd DEBUG
  //double t6 = MPI_Wtime();
  //printf("bigFFT.backward time:  mype = %i, t3 time = %0.7f sec\n",mype,t6-t5);

  // rearrange data so that y will be distributed and x will end up
  // local after alltoallv i.e. current local index (y) must be
  // varying slowest in sent array
  for (int i=0; i<np0loc; i++)
    for (int k=0; k<np2loc; k++)
      for (int j=0; j<np1; j++)
      {
        const FftDataType v = bigdata->fftout[j+i*np1+k*np1*np0loc];
        bigdata->mpibuf[i+k*np0loc+j*np2loc*np0loc] = v;
      }

  int scnts[nprow],rcnts[nprow],sdisp[nprow],rdisp[nprow];
  soffset = 0;
  roffset = 0;
  for (int i=0; i<nprow; i++)
  {
    int sendblk = 2*np0loc*bigdata->np1loc_row_[i]*np2loc;
    int recvblk = 2*bigdata->np0loc_row_[i]*np1locf*np2loc;
    scnts[i] = sendblk;
    rcnts[i] = recvblk;
    sdisp[i] = soffset;
    rdisp[i] = roffset;
    soffset += sendblk;
    roffset += recvblk;
  }

  status = MPI_Alltoallv((FftCommType*)&bigdata->mpibuf[0],&scnts[0],&sdisp[0],FFT_COMM_TYPE,
                         (FftCommType*)&bigdata->pack[0],&rcnts[0],&rdisp[0],FFT_COMM_TYPE,bigdata->colcomm_);
  assert ( status == 0 );

  //ewd DEBUG
  //double t7 = MPI_Wtime();
  //printf("bigFFT.backward time:  mype = %i, t7 time = %0.7f sec, alltoallv bwd2\n",mype,t7-t6);


  // reorder data so that x varies fastest for x-transform
  for (int k=0; k<np2loc; k++)
    for (int j=0; j<np1locf; j++)
    {
      int i = 0;
      int recvstart = 0;
      for (int tr=0; tr<nprow; tr++)
      {
        for (int iloc=0; iloc<bigdata->np0loc_row_[tr]; iloc++)
        {
          const FftDataType v = bigdata->pack[recvstart + k*bigdata->np0loc_row_[tr] + j*bigdata->np0loc_row_[tr]*np2loc + iloc];

          if (i == 0 || i == np0/2 )
            bigdata->fftin[i+j*np0+k*np0*np1locf] = v;
          else if (i < np0/2)
          {
            bigdata->fftin[i+j*np0+k*np0*np1locf] = v;
            bigdata->fftin[np0-i+j*np0+k*np0*np1locf] = conj(v);
          }
          i++;
        }
        recvstart += bigdata->np0loc_row_[tr]*np1locf*np2loc;
      }
    }

  inc2 = np0;
  int nxt = np2loc*np1locf;
#ifdef USE_FFTW
  fftw(bigdata->bwplan0,nxt,(FFTW_COMPLEX*)&bigdata->fftin[0],inc1,inc2,
       (FFTW_COMPLEX*)&bigdata->fftout[0],inc1,inc2);
#endif
#ifdef USE_ESSL
#ifdef USE_SINGLEPREC_FFT
  scft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np0,nxt,isign,esslscale,&bigdata->aux1xb[0],bigdata->naux1x,&bigdata->aux2[0],bigdata->naux2);
#else
  dcft(initflag,&bigdata->fftin[0],inc1,inc2,&bigdata->fftout[0],inc1,inc2,
        np0,nxt,isign,esslscale,&bigdata->aux1xb[0],bigdata->naux1x,&bigdata->aux2[0],bigdata->naux2);
#endif  
#endif  
  //ewd DEBUG
  //double t8 = MPI_Wtime();
  //printf("bigFFT.backward time:  mype = %i, t8 time = %0.7f sec\n",mype,t8-t7);

  for (int i=0; i<np0*np1locf*np2loc; i++)
    rhok[i] = (double)(scale*creal(bigdata->fftout[i]));

  return 0;
}
////////////////////////////////////////////////////////////////////////////////
int bigfft_np0loc_bwd(BIGFFT_DATA *bigdata) { return bigdata->np0loc_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_np1loc_fwd(BIGFFT_DATA *bigdata) { return bigdata->np1locf_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_np1loc_bwd(BIGFFT_DATA *bigdata) { return bigdata->np1locb_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_np2loc_fwd(BIGFFT_DATA *bigdata) { return bigdata->np2loc_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_np0loc_row(BIGFFT_DATA *bigdata, int irow) { return bigdata->np0loc_row_[irow]; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_np1loc_row(BIGFFT_DATA *bigdata, int irow) { return bigdata->np1loc_row_[irow]; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_np1loc_col(BIGFFT_DATA *bigdata, int icol) { return bigdata->np1loc_col_[icol]; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_np2loc_col(BIGFFT_DATA *bigdata, int icol) { return bigdata->np2loc_col_[icol]; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_origin_np0_bwd(BIGFFT_DATA *bigdata) 
{ 
  int sum = 0;
  for (int i=0; i<bigdata->myrow_; i++)
    sum += bigdata->np0loc_row_[i];
  return sum;
}
////////////////////////////////////////////////////////////////////////////////
int bigfft_origin_np1_bwd(BIGFFT_DATA *bigdata) 
{ 
  int sum = 0;
  for (int i=0; i<bigdata->mycol_; i++)
    sum += bigdata->np1loc_col_[i];
  return sum;
}
////////////////////////////////////////////////////////////////////////////////
int bigfft_origin_np1_fwd(BIGFFT_DATA *bigdata) 
{ 
  int sum = 0;
  for (int i=0; i<bigdata->myrow_; i++)
    sum += bigdata->np1loc_row_[i];
  return sum;
}
////////////////////////////////////////////////////////////////////////////////
int bigfft_origin_np2_fwd(BIGFFT_DATA *bigdata) 
{ 
  int sum = 0;
  for (int i=0; i<bigdata->mycol_; i++)
    sum += bigdata->np2loc_col_[i];
  return sum;
}
////////////////////////////////////////////////////////////////////////////////
int bigfft_nprow(BIGFFT_DATA *bigdata) { return bigdata->nprow_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_npcol(BIGFFT_DATA *bigdata) { return bigdata->npcol_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_myrow(BIGFFT_DATA *bigdata) { return bigdata->myrow_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_mype(BIGFFT_DATA *bigdata) { return bigdata->mype_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_mycol(BIGFFT_DATA *bigdata) { return bigdata->mycol_; }
////////////////////////////////////////////////////////////////////////////////
MPI_Comm bigfft_comm(BIGFFT_DATA *bigdata) { return bigdata->comm_; }
////////////////////////////////////////////////////////////////////////////////
int bigfft_gridpe_fwd(BIGFFT_DATA *bigdata, int x, int y, int z) 
{
  int row = bigdata->iprow_y_[y];
  int col = bigdata->ipcol_z_[z];
  int pe = row + col*bigdata->nprow_;
  return pe;
}
////////////////////////////////////////////////////////////////////////////////
